#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>

#include "matrix.h"

static void exitif(const char *err, bool x) {
  if (!x) {
    perror(err);
    exit(1);
  }
}

void readpnm(FILE *f, Matrix<float> *channels) {
  int format, xsize, ysize;
  int maxval = 1;
  if(fscanf(f, "P%d %d %d", &format, &xsize, &ysize)<3 ||
     format < 1 || format > 6 ||
     ( (format%3)!=1 && !fscanf(f, "%d",&maxval)))
  {
    fprintf(stderr,"solve: Not a PPM file\n");
    exit(1);
  }
  
  getc(f);

  channels[0] = Matrix<float>(xsize, ysize);
  channels[1] = Matrix<float>(xsize, ysize);
  channels[2] = Matrix<float>(xsize, ysize);

  int bit = 0; 
  int data = 0;
  for(int y=0; y<ysize; y++) {
    for(int x=0; x<xsize; x++) {
      int r, g, b;
      switch(format) {
	case 1:
	case 2:
	  exitif("fscanf", 1 == fscanf(f,"%d",&g));
	  r = b = g;
	  break;

	case 3:
	  exitif("fscanf", 3 == fscanf(f,"%d %d %d",&r, &g, &b));
	  break;

	case 4:
	  if(!bit) { data=getc(f); bit=8; }
	  r = b = g = !!(data & 128);
	  data<<=1;
	  bit--;
	  break;

	case 5:
	  r = b = g = getc(f);
	  break;

	case 6:
	  r = getc(f);
	  b = getc(f);
	  g = getc(f);
      }
      channels[0].set(x, y, pow(r / (float)maxval, 2.2));
      channels[1].set(x, y, pow(g / (float)maxval, 2.2));
      channels[2].set(x, y, pow(b / (float)maxval, 2.2));
    }
  }
}

Matrix<float> readpgm(FILE *f, int reduce) {
  int format, xsize, ysize;
  int maxval = 1;
  if(fscanf(f, "P%d %d %d", &format, &xsize, &ysize)<3 ||
     format < 1 || format > 6 ||
     ( (format%3)!=1 && !fscanf(f, "%d",&maxval)))
  {
    fprintf(stderr,"solve: Not a PPM file\n");
    exit(1);
  }
  
  getc(f);

  Matrix<float> ret = Matrix<float>((xsize + reduce - 1) / reduce,
				    (ysize + reduce - 1) / reduce);
  float mult = 1.0 / (reduce * reduce);

  int bit = 0; 
  int data = 0;
  for(int y=0; y<ysize; y++) {
    for(int x=0; x<xsize; x++) {
      int r, g, b;
      switch(format) {
	case 1:
	case 2:
	  exitif("fscanf", 1 == fscanf(f,"%d",&g));
	  r = b = g;
	  break;

	case 3:
	  exitif("fscanf", 3 == fscanf(f,"%d %d %d",&r, &g, &b));
	  break;

	case 4:
	  if(!bit) { data=getc(f); bit=8; }
	  r = b = g = !!(data & 128);
	  data<<=1;
	  bit--;
	  break;

	case 5:
	  r = b = g = getc(f);
	  break;

	case 6:
	  r = getc(f);
	  b = getc(f);
	  g = getc(f);
      }
      ret.set(x/reduce, y/reduce, ret.get(x/reduce, y/reduce) + mult * pow((r + g + b) / 3.0 / maxval, 2.2));
    }
  }
  return ret;
}

std::string filename_for_image(int image) {
  char tmp[256];
  sprintf(tmp, "images/light_%d.jpg", image);
  return std::string(tmp);
}

Matrix<float> read_image(int image) {
  char tmp[512];
  sprintf(tmp, "djpeg -grayscale -pnm %s |  pnmcut -left 500 -right -500 -top 20 -bottom -350", filename_for_image(image).c_str());
  fprintf(stderr, "Running: %s\n", tmp);
  FILE *f = popen(tmp, "r");
  Matrix<float> ret = readpgm(f, 4);
  fclose(f);
  return ret;
}

void read_full_image(int image, Matrix<float> *channels) {
  char tmp[512];
  sprintf(tmp, "djpeg -pnm %s", filename_for_image(image).c_str());
  fprintf(stderr, "Running: %s\n", tmp);
  FILE *f = popen(tmp, "r");
  readpnm(f, channels);
  fclose(f);
}

void write_pnm(FILE *f, Matrix<float> *channels) {
  int xsize = channels->columns();
  int ysize = channels->rows();
  fprintf(f, "P6\n%d %d\n255\n", xsize, ysize);
  for (int y = 0; y < ysize; y++) {
    for (int x = 0; x < xsize; x++) {
      for (int c = 0; c < 3; c++) {
	float v = channels[c].get(x, y);
	v = pow(v, 1.0/2.2);
	int iv = (int)(v * 255);
	if (iv < 0) iv = 0;
	if (iv > 255) iv = 255;
	putc(iv, f);
      }
    }
  }
}

void write_pgm(FILE *f, Matrix<float> m) {
  int xsize = m.columns();
  int ysize = m.rows();
  fprintf(f, "P5\n%d %d\n255\n", xsize, ysize);
  for (int y = 0; y < ysize; y++) {
    for (int x = 0; x < xsize; x++) {
      float v = m.get(x, y);
      int iv = (int)(v * 255);
      if (iv < 0) iv = 0;
      if (iv > 255) iv = 255;
      putc(iv, f);
    }
  }
}

