#include "lightbuffer.h"
#include "lights.h"
#include "timing.h"

#include <pthread.h>
#include <sys/time.h>

Color black_;

// restart handling for write()
ssize_t WRITE(int fd, const void* buf, size_t count) {
  ssize_t written, ret;
  written = 0;
  while (written < count) {
    ret = write(fd, ((const char*)buf) + written, count - written);
    if (ret <= 0) {
      if (errno == EINTR) continue;
      if (written) return written;
      return ret;
    }
    written += ret;
  }
  return written;
}

class LightRow {
public:
  LightRow(Color* colors, int size) : colors_(colors), size_(size) {}
  Color& operator[](int col) const {
    if (col < 0 || col >= size_) {
      black_ = Color();
      return black_;
    }
    return colors_[col];
  }
private:
  Color* colors_;
  int size_;
};

class LightGrid {
public:
  LightGrid() : colors_(NULL), num_rows_(0) {}
  LightGrid(Color** colors, int num_rows) :
    colors_(colors),
    num_rows_(num_rows) {
  }
  LightRow operator[](int row) const {
    if (row < 0) return LightRow(colors_[0], 0);
    if (row >= num_rows_) return LightRow(colors_[0], 0);
    return LightRow(colors_[row], colors_[row + 1] - colors_[row]);
  }

  LightGrid operator+(int rows) const {
    return LightGrid(colors_ + rows, num_rows_ - rows);
  }
  
private:
  Color** colors_;
  int num_rows_;
};

class Output {
public:
  virtual ~Output() {}
  virtual void GetGeometry(vector<int>* row_lengths) = 0;
  virtual void Prepare(LightGrid l) = 0;
  virtual void Show(LightGrid l) = 0;
  virtual void Sync() {}
};

class OutputLW : public Output {
public:
  OutputLW(const string &device,
	   int num_strings) : num_strings_(num_strings),
			      tw_(device, num_strings) {
  }
  virtual void GetGeometry(vector<int>* row_lengths) {
    for (int i = 0; i < num_strings_; i++) {
      row_lengths->push_back(50);
    }
  }
  virtual void Prepare(LightGrid lights) {}
  virtual void Show(LightGrid lights) {
    vector<LampColor> c(num_strings_);
    for (int i = 0; i < 50; i++) {
      for (int s = 0; s < num_strings_; s++) {
	int x = i + s*50;
	c[s].intensity = 0xFF;
	c[s].color[0] = lights[s][i].color[0];
	c[s].color[1] = lights[s][i].color[1];
	c[s].color[2] = lights[s][i].color[2];
      }
      tw_.SetPixels(i, c);
    }
  }

private:
  int num_strings_;
  TeensyLightWallRelay tw_;
};

// rotates 8x8 bytes bitwise
unsigned long long bitrot(unsigned long long i) {
  unsigned long long a, b, c, d;
  a = i & 0xAA00AA00AA00AA00ULL;
  b = i & 0x5500550055005500ULL;
  c = i & 0x00AA00AA00AA00AAULL;
  d = i & 0x0055005500550055ULL;
#define R(N) ((a >> (8*N)) | (b << (N)) | (c >> (N)) | (d << (8*N)))
#define L(N) ((a >> (N)) | (b >> (8*N)) | (c << (8*N)) | (d << (N)))
  i = R(1);
  a = i & 0xCCCC0000CCCC0000ULL;
  b = i & 0x3333000033330000ULL;
  c = i & 0x0000CCCC0000CCCCULL;
  d = i & 0x0000333300003333ULL;
  i = R(2);
  a = i & 0xF0F0F0F000000000ULL;
  b = i & 0x0F0F0F0F00000000ULL;
  c = i & 0x00000000F0F0F0F0ULL;
  d = i & 0x000000000F0F0F0FULL;
  i = R(4);
  return i;
}

static int geom_max(const std::vector<int>& geom) {
  int ret = 0;
  for (int i = 0; i < geom.size(); i++) {
    ret = std::max(ret, geom[i]);
  }
  fprintf(stderr, "MAX = %d\n", ret);
  return ret;
}

static int geom_min(const std::vector<int>& geom) {
  int ret = geom[0];
  for (int i = 1; i < geom.size(); i++) {
    ret = std::min(ret, geom[i]);
  }
  return ret;
}

class OutputOcto : public Output {
public:
  OutputOcto(const string& device,
	     const std::vector<int>& geometry,
	     int byte_order) :
    geometry_(geometry),
    lights_per_string_(geom_max(geometry)),
    bytes_(geom_max(geometry) * 8 * 3 + 3),
    byte_order_(byte_order),
    fd_(open(device.c_str(), O_RDWR)) {
    
    if (fd_ < 0) {
      string err = "Failed to open " + device;
      perror(err.c_str());
      exit(1);
    }

    struct termios term_settings;
    if (tcgetattr(fd_, &term_settings) < 0) {
      perror("tcgetattr");
      exit(1);
    }
    cfmakeraw(&term_settings);
    cfsetspeed(&term_settings, B230400);
    term_settings.c_cflag &=~ CSTOPB;
    if (tcsetattr(fd_, TCSADRAIN, &term_settings) < 0) {
      perror("tcsetattr");
      exit(1);
    }
  }

  virtual void GetGeometry(vector<int>* row_lengths) {
    for (int i = 0; i < 8; i++) {
      row_lengths->push_back(geometry_[i]);
    }
  }

  virtual void Prepare(LightGrid g) {
    bytes_[0] = 0xee;
    bytes_[1] = (lights_per_string_ >> 8) & 0xff;
    bytes_[2] = lights_per_string_ & 0xff;
    int j = 3;
    int shortest_string = geom_min(geometry_);
    for (int i = 0; i < lights_per_string_; i++) {
      for (int component = 0; component < 3; component++) {
	int c = (byte_order_ >> (component * 4)) & 0xf;
#if 0
	for (int bit = 0; bit < 8; bit++) {
	  unsigned char tmp = 0;
	  for (int s = 0; s < 8; s++) {
	    tmp <<= 1;
	    if (i < geometry_[s]) {
	      tmp |= (g[s][i].color[c] >> (7 - bit)) & 1;
	    }
	  }
	  bytes_[j++] = tmp;
	}
#else
	unsigned long long tmp = 0;
	if (i < shortest_string) {
	  for (int s = 0; s < 8; s++) {
	    tmp <<= 8;
	    tmp |= g[s][i].color[c];
	  }
	} else {
	  for (int s = 0; s < 8; s++) {
	    tmp <<= 8;
	    if (i < geometry_[s]) {
	      tmp |= g[s][i].color[c];
	    }
	  }
	}
	//fprintf(stderr, "{%d, %d, %d}\n", c.color[0], c.color[1], c.color[2]);
	tmp = bitrot(tmp);
//	fprintf(stderr, "L: %llx\n", tmp);
	for (int s = 0; s < 8; s++) {
	  bytes_[j++] = tmp;
	  tmp >>= 8;
	}
#endif
      }
    }
    if (WRITE(fd_, bytes_.data() ,bytes_.size() - 1) != bytes_.size() -1) {
      perror("Write to Octo.");
      exit(1);
    }
  }
  virtual void Show(LightGrid g) {
    if (WRITE(fd_, bytes_.data() + bytes_.size() - 1, 1) != 1) {
      perror("Write to Octo.");
      exit(1);
    }
  }
  virtual void Sync() {
    unsigned char sync[2];
    sync[0] = 0xED;
    sync[1] = syncnum_++;
    if (WRITE(fd_, sync, 2) != 2) {
      perror("Write to Octo.");
      exit(1);
    }
    do {
      if (read(fd_, sync, 1) != 1) {
	perror("Write to Octo");
	exit(1);
      }
    } while(sync[0] != sync[1]);
  }

private:
  std::vector<unsigned char> bytes_;
  std::vector<int> geometry_;
  int lights_per_string_;
  int byte_order_;
  int fd_;
  unsigned char syncnum_;
};

class StdOut : public Output {
public:
  StdOut(std::vector<int> geometry) : geometry_(geometry) {}

  virtual void GetGeometry(vector<int>* row_lengths) {
    for (int i = 0; i < geometry_.size(); i++) {
      row_lengths->push_back(geometry_[i]);
    }
  }
  virtual void Prepare(LightGrid lights) {
  }
  virtual void Show(LightGrid lights) {
    putchar(0xEE);
    for (int s = 0; s < geometry_.size(); s++) {
      putchar(geometry_[s] >> 8);
      putchar(geometry_[s]);
      for (int l = 0; l < geometry_[s]; l++) {
	putchar(lights[s][l].color[0]);
	putchar(lights[s][l].color[1]);
	putchar(lights[s][l].color[2]);
      }
    }
    // Terminator
    putchar(0x00);
    putchar(0x00);
  }

private:
  std::vector<int> geometry_;
};

class OutputAnsi : public Output {
public:
  OutputAnsi(std::vector<int> geometry) : geometry_(geometry) {
    printf("\033[2J");
  }
  virtual void GetGeometry(vector<int>* row_lengths) {
    for (int i = 0; i < geometry_.size(); i++) {
      row_lengths->push_back(geometry_[i]);
    }
  }
  virtual void Prepare(LightGrid lights) {
  }
  virtual void Show(LightGrid lights) {
    printf("\033[H");
    for (int s = 0; s < geometry_.size(); s++) {
      for (int l = 0; l < geometry_[s]; l++) {
	printf("\033[48;2;%d;%d;%dm ", lights[s][l].color[0], lights[s][l].color[1], lights[s][l].color[2]);
	if ((l % 100) == 99) printf("\r\n");
      }
      printf("\r\n");
    }
    printf("\033[48;2;255;255;255m\033[0m");
    fflush(stdout);
    usleep(20000);
  }

private:
  std::vector<int> geometry_;
};

class MultiOut : public Output {
public:
  MultiOut() {}
  ~MultiOut() {
    for (int i = 0; i < outputs_.size(); i++) {
      delete outputs_[i];
    }
  }

  void AddOutput(Output* out) {
    outputs_.push_back(out);
    vector<int> geometry;
    out->GetGeometry(&geometry);
    rows_.push_back(geometry.size());
  }

  virtual void GetGeometry(vector<int>* row_lengths) {
    for (int i = 0; i < outputs_.size(); i++) {
      outputs_[i]->GetGeometry(row_lengths);
    }
  }
  virtual void Prepare(LightGrid g) {
    int row = 0;
    for (int i = 0; i < outputs_.size(); i++) {
      outputs_[i]->Prepare(g + row);
      row += rows_[i];
    }
  }
  struct ThreadData {
    pthread_t thread;
    Output* output;
    LightGrid g;
  };
  static void* Run(void* arg) {
    ThreadData* thread = (ThreadData*)arg;
    thread->output->Show(thread->g);
  }
  // TODO Synchronize better!!
  virtual void Show(LightGrid g) {
    int row = 0;
    std::vector<ThreadData> threads(outputs_.size());
    for (int i = 0; i < outputs_.size(); i++) {
      threads[i].g = g + row;
      threads[i].output = outputs_[i];
      pthread_create(&threads[i].thread, NULL, &MultiOut::Run, &threads[i]);
      row += rows_[i];
    }
    for (int i = 0; i < outputs_.size(); i++) {
      pthread_join(threads[i].thread, NULL);
    }
  }

  vector<Output*> outputs_;
  vector<int> rows_;
  int num_lights_;
};

class LightBufferImpl : public LightBuffer {
public:
  LightBufferImpl(Output* output) : output_(output), frame_(0), max_frames_(0x7fffffff) {
    start_ = Time();
    std::vector<int> geometry;
    output->GetGeometry(&geometry);
    int total = 0;
    for (int i = 0; i < geometry.size(); i++) {
      total += geometry[i];
    }
    colors_.resize(total);
    Color* tmp = colors_.data();
    for (int i = 0; i < geometry.size(); i++) {
      rows_.push_back(tmp);
      tmp += geometry[i];
    }
    rows_.push_back(tmp);
    if (getenv("MAX_FRAMES")) {
      max_frames_ = atoi(getenv("MAX_FRAMES"));
    }
  }
  ~LightBufferImpl() {
    delete output_;
  }

  virtual void GetGeometry(std::vector<int>* row_lengths) {
    output_->GetGeometry(row_lengths);
  }
  virtual int Set(int strip, int light, Color c) {
    LightGrid(rows_.data(), rows_.size())[strip][light] = c;
  }
  virtual void Show() {
    output_->Prepare(LightGrid(rows_.data(), rows_.size() - 1));
    output_->Show(LightGrid(rows_.data(), rows_.size() - 1));
    struct timeval now_;
    frame_++;
    now_ = Time();
    double fps = frame_ / (now_ - start_);
    fprintf(stderr, "\r Frame: %d   FPS: %6.3f   ", frame_, fps);
    if (frame_ > max_frames_) {
      fprintf(stderr, "\n");
      exit(0);
    }
  }

  virtual void Sync() {
    output_->Sync();
  }
  
private:
  Output* output_;
  std::vector<Color> colors_;
  std::vector<Color*> rows_;
  int frame_;
  int max_frames_;
  struct timeval start_;
};


LightBuffer* GetLightBuffer() {
  Output* out;
  vector<int> geometry;
  if (getenv("LIGHTBUFFER_ANSI")) {
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    out = new OutputAnsi(geometry);
  } else {
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    geometry.push_back(450);
    MultiOut* ret = new MultiOut();
    ret->AddOutput(new OutputOcto("/dev/ttyACM0", geometry, 0x210));
    ret->AddOutput(new OutputOcto("/dev/ttyACM1", geometry, 0x210));
    out = ret;
  }
  return new LightBufferImpl(out);
}

LightBuffer* GetLightBufferStdout() {
  std::vector<int> geometry;
  geometry.push_back(450);
  geometry.push_back(450);
  geometry.push_back(450);
  geometry.push_back(450);

  geometry.push_back(450);
  geometry.push_back(450);
  geometry.push_back(450);
  geometry.push_back(450);

  geometry.push_back(450);
  geometry.push_back(450);
  geometry.push_back(450);
  geometry.push_back(450);

  geometry.push_back(450);
  geometry.push_back(450);
  geometry.push_back(450);
  geometry.push_back(450);

  return new LightBufferImpl(new StdOut(geometry));
}
