#ifndef LIGHTS_H
#define LIGHTS_H

#include <sys/select.h>
#include <sys/time.h>
#include <termios.h>
#include <string>
#include <vector>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>

using namespace std;

// Number of ports used to control lights.
#define NUM_PORTS 2
#define MAX_STRINGS (8*NUM_PORTS)
#define PACKET_SIZE (1 + NUM_PORTS * 26)

struct LampColor {
  LampColor() { clear(); }
  void clear() {
    intensity = 0;
    color[0] = 0;
    color[1] = 0;
    color[2] = 0;
  }
  unsigned char intensity;
  // color[0] = r, color[1] = g, color[2] = b
  // Only top 4 bits are actually used for r/g/b
  unsigned char color[3];
};

static bool Write(int fd, const char *s, int n) {
  int written = 0;
  while (written < n) {
    int tmp = write(fd, s + written, n - written);
    if (tmp < 0 && errno == EINTR) continue;
    if (tmp < 0) return false;
    written += tmp;
  }
  return true;
}

class TeensyLightWallRelay {
public:
 TeensyLightWallRelay(const string &device, int num_strings,
		      bool monitor=false) :
   device_(device),
   num_strings_(num_strings)
  {
    if (device_ != "") {
      fd_ = open(device_.c_str(),  O_RDWR);
      if (fd_ < 0) {
	string err = "Failed to open " + device_;
	perror(err.c_str());
	exit(1);
      }
      struct termios term_settings;
      if (tcgetattr(fd_, &term_settings) < 0) {
	perror("tcgetattr");
	exit(1);
      }
      cfmakeraw(&term_settings);
      cfsetspeed(&term_settings, B230400);
      term_settings.c_cflag &=~ CSTOPB;
      if (tcsetattr(fd_, TCSADRAIN, &term_settings) < 0) {
	perror("tcsetattr");
	exit(1);
      }
      if (monitor && !fork()) {
	while (1) {
	  char buf[1024];
	  int bytes = read(fd_, buf, 1);
	  if (bytes > 0) {
	    Write(2, buf, bytes);
	  } else {
	    if (errno == EINTR) continue;
	    Write(2, "\nClosed", 7);
	    exit(0);
	  }
	}
      }
    } else {
      // stdout
      fd_ = 1;
    }
  }


  void EncodePixels(int lamp, const vector<LampColor> &colors) {
    buffer_[0] = 0xEE; // sync
    for (int i = 1; i < PACKET_SIZE; i++) buffer_[i] = 0;
    for (int i = 0; i < colors.size() && i < MAX_STRINGS; i++) {
      int bit = 1;
      bit += i / 8;
      int mask = 1 << (i & 7);
      for (int lamp_bit = 32; lamp_bit; lamp_bit >>= 1) {
	if (lamp_bit & lamp) buffer_[bit] |= mask;
	bit += NUM_PORTS;
      }
      for (int intensity_bit = 128; intensity_bit; intensity_bit >>= 1) {
	if (colors[i].intensity & intensity_bit) {
	  buffer_[bit] |= mask;
	}
	bit += NUM_PORTS;
      }
      for (int color = 0; color < 3; color++) {
	for (int color_bit = 128; color_bit >= 16; color_bit >>= 1) {
	  if (colors[i].color[2-color] & color_bit) {
	    buffer_[bit] |= mask;
	  }
	  bit += NUM_PORTS;
	}
      }
    }
  }

#if 1
  void SetPixels(int lamp, const vector<LampColor> &colors) {
    EncodePixels(lamp, colors);

    if (write(fd_, buffer_, PACKET_SIZE) != PACKET_SIZE) {
      perror("Write to Teensy.");
      exit(1);
    }
  }
#else
  void SetPixels(int lamp, const vector<LampColor> &colors) {
    int bufpos_ = 0;
    buffer_[bufpos_++] = 0xEE; // sync
    for (int lamp_bit = 32; lamp_bit; lamp_bit >>= 1) {
      for (int p = 0; p < NUM_PORTS; p++) {
	buffer_[bufpos_++] = (lamp_bit & lamp) ? 0xFF : 0x00;
      }
    }
    for (int intensity_bit = 128; intensity_bit; intensity_bit >>= 1) {
      unsigned char tmp = 0;
      for (int i = 0; i < colors.size(); i++) {
	if (colors[i].intensity & intensity_bit) {
	  tmp |= 1 << i;
	}
	if ((i & 7) == 7 || i == colors.size()) {
	  buffer_[bufpos_++] = tmp;
	  tmp = 0;
	}
      }
    }
    for (int color = 0; color < 3; color++) {
      for (int color_bit = 128; color_bit >= 16; color_bit >>= 1) {
	unsigned char tmp = 0;
	for (int i = 0; i < colors.size(); i++) {
 	  if (colors[i].color[2-color] & color_bit) {
	    tmp |= 1 << i;
	  }
	  if ((i & 7) == 7 || i == colors.size()) {
	    buffer_[bufpos_++] = tmp;
	    tmp = 0;
	  }
	}
      }
    }
    if (write(fd_, buffer_, bufpos_) != bufpos_) {
      perror("Write to Teensy.");
      exit(1);
    }
  }
#endif

  bool ReadPacket(int fd) {
    if (read(fd, buffer_, PACKET_SIZE) != PACKET_SIZE) {
      return false;
    }
    if (buffer_[0] != 0xEE) {
      fprintf(stderr, "First byte not EE\n");
      exit(1);
    }
    return true;
  }

  bool DecodePixels(int *lamp, vector<LampColor> *colors) {
    *lamp = 0;
    for (int i = 0; i < colors->size(); i++) (*colors)[i].clear();
    for (int i = 0; i < colors->size() && i < MAX_STRINGS; i++) {
      int bit = 1;
      bit += i / 8;
      int mask = 1 << (i & 7);
      for (int lamp_bit = 32; lamp_bit; lamp_bit >>= 1) {
	if (buffer_[bit] & mask) *lamp |= lamp_bit;
	bit += NUM_PORTS;
      }
      for (int intensity_bit = 128; intensity_bit; intensity_bit >>= 1) {
	if (buffer_[bit] & mask) {
	  (*colors)[i].intensity |= intensity_bit;
	}
	bit += NUM_PORTS;
      }
      for (int color = 0; color < 3; color++) {
	for (int color_bit = 128; color_bit >= 16; color_bit >>= 1) {
	  if (buffer_[bit] & mask) {
	    (*colors)[i].color[2-color] |= color_bit;
	  }
	  bit += NUM_PORTS;
	}
      }
    }
  }
  
  // Decode a packet.
  bool GetPixels(int fd, int *lamp, vector<LampColor> *colors) {
    if (!ReadPacket(fd)) return false;
    DecodePixels(lamp, colors);
    return true;
  }

  void SendFile(int fd) {
    while (1) {
      if (!ReadPacket(fd)) return;
      if (write(fd_, buffer_, PACKET_SIZE) != PACKET_SIZE) {
	perror("Write to Teensy.");
	exit(1);
      }
      fprintf(stderr,".");
      fflush(stderr);
    }
  }

  void test() {
    vector<LampColor> colors(MAX_STRINGS);
    fprintf(stderr, "TEST\n");
    colors[0].intensity = 0xC7;
    colors[0].color[0] = 0x20;
    colors[0].color[1] = 0x30;
    colors[0].color[2] = 0x40;
    EncodePixels(17, colors);
    colors[0].intensity = 0x0;
    colors[0].color[0] = 0x0;
    colors[0].color[1] = 0x0;
    colors[0].color[2] = 0x0;
    int lamp;
    DecodePixels(&lamp, &colors);
    fprintf(stderr, "c0.intensity = %x\n", colors[0].intensity);
    fprintf(stderr, "c0.color[0] = %x\n", colors[0].color[0]);
    fprintf(stderr, "c0.color[1] = %x\n", colors[0].color[1]);
    fprintf(stderr, "c0.color[2] = %x\n", colors[0].color[2]);
  }

  void Init() {
    if (fd_ == 1) {
      // No init for stdout
      return;
    }
    vector<LampColor> colors(MAX_STRINGS);
    for (int i = 0; i < colors.size(); i++) {
      colors[i].intensity = 0xCC;
      colors[i].color[0] = 0xFF;
      colors[i].color[1] = 0xFF;
      colors[i].color[2] = 0xFF;
    }
    for (int lamp = 0; lamp <= 62; lamp++) {
      SetPixels(lamp, colors);
      struct timeval timeout;
      timeout.tv_sec = 0;
      timeout.tv_usec = 25000;
      fprintf(stderr,".");
      fflush(stderr);
      select(0, NULL, NULL, NULL, &timeout);
    }
    fprintf(stderr, "\n");
  }


protected:
  string device_;
  int num_strings_;
  int fd_;
  unsigned char buffer_[PACKET_SIZE];
};

#endif
