/*
 *   nnls.c  (c) 2002-2009 Turku PET Centre
 *   This file contains the routine NNLS (nonnegative least squares)
 *   and the subroutines required by it, except h12, which is in
 *   file 'lss_h12.c'.
 *   
 *  This routine is based on the text and fortran code in
 *  C.L. Lawson and R.J. Hanson, Solving Least Squares Problems,
 *  Prentice-Hall, Englewood Cliffs, New Jersey, 1974.
 *  Version:
 *  2002-08-19 Vesa Oikonen
 *  2003-05-08 Kaisa Sederholm & VO
 *  Included function nnlsWght().
 *  2003-05-12 KS
 *  Variable a_dim1 excluded
 *  
 *  Usage of the coefficient matrix altered so that it is
 *  given in a[][] instead of a[].
 *  2003-11-06 VO
 *  If n<2, then itmax is set to n*n instead of previous n*3.
 *  2004-09-17 VO
 *  Doxygen style comments.
 *  2006-24-04 Pauli Sundberg
 *  Added some debuging output, and made some comments more precise.
 *  2007-05-17 VO
 *  2009-04-16 VO
 *  Corrected a bug in nnls() which may have caused an infinite loop.
 *  2009-04-27 VO
 *  Added function nnlsWghtSquared() for faster pixel-by-pixel calculations.
 *  Checking for exceeding iteration count is corrected in nnls().
 */ 
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <pthread.h>

// #define CUBLAS

#ifdef CUBLAS
#include <cublas.h>

pthread_mutex_t handle_mutex = PTHREAD_MUTEX_INITIALIZER;
static int inited = 0;

void InitCublas() {
  pthread_mutex_lock(&handle_mutex);
  if (!inited) {
    inited = 1;
    cublasStatus_t status = cublasInit();
    if (status != CUBLAS_STATUS_SUCCESS) {
      fprintf(stderr, "Cublas init failed: %d\n", status);
      exit(1);
    }
  }
  pthread_mutex_unlock(&handle_mutex);
}

#endif

static inline float MAX(float a, float b) { return a >= b ? a : b; }
static inline float ABS(float x) { return x >= 0 ? x : -x; }
static inline float sqr(float x) { return x * x; }

#include <smmintrin.h>
#include <immintrin.h>

//#undef __AVX__

float dot(float *a, float *b, size_t count) {
  float ret = 0.0;

#ifdef CUBLAS
  InitCublas();
  float *tmp_a, *tmp_b;
  cublasStatus_t status;
  status = cublasAlloc(count, sizeof(float), (void**)&tmp_a);
  if (status != CUBLAS_STATUS_SUCCESS) {
    fprintf(stderr, "cublasAlloc failed, %d\n", status);
    exit(1);
  }
  status = cublasSetVector(count, sizeof(float), a, 1, tmp_a, 1);
  if (status != CUBLAS_STATUS_SUCCESS) {
    fprintf(stderr, "cublasSetVector failed, %d\n", status);
    exit(1);
  }
  status = cublasAlloc(count, sizeof(float), (void**)&tmp_b);
  if (status != CUBLAS_STATUS_SUCCESS) {
    fprintf(stderr, "cublasAlloc failed, %d\n", status);
    exit(1);
  }
  status = cublasSetVector(count, sizeof(float), b, 1, tmp_b, 1);
  if (status != CUBLAS_STATUS_SUCCESS) {
    fprintf(stderr, "cublasSetVector failed, %d\n", status);
    exit(1);
  }
  ret = cublasSdot(count, tmp_a, 1, tmp_b, 1);
  status = cublasGetError();
  if (status != CUBLAS_STATUS_SUCCESS) {
    fprintf(stderr, "cublasSdot failed, %d\n", status);
    exit(1);
  }
  cublasFree(tmp_a);
  cublasFree(tmp_b);
#else
#if 0
  __m256 tmp = _mm256_set_ps(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);
  while (count >= 8) {
    tmp = _mm256_add_ps(
      tmp,
      _mm256_mul_ps(_mm256_set_ps(a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7]),
                    _mm256_set_ps(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7])));

    a += 8;
    b += 8;
    count -= 8;
  }
  ret = tmp[0] + tmp[1] + tmp[2] + tmp[3] + tmp[4] + tmp[5] + tmp[6] + tmp[7];
  while (count > 0) {
    ret += *(a++) * *(b++);
    count--;
  }
#endif
#if 1
  float tmp2 = 0.0;
  while (count >= 8) {
    ret += _mm_cvtss_f32(_mm_dp_ps(
			   _mm_set_ps(a[0],a[1],a[2],a[3]),
			   _mm_set_ps(b[0],b[1],b[2],b[3]), 0xff));
    tmp2 += _mm_cvtss_f32(_mm_dp_ps(
			    _mm_set_ps(a[4],a[5],a[6],a[7]),
			    _mm_set_ps(b[4],b[5],b[6],b[7]), 0xff));
    a += 8;
    b += 8;
    count -= 8;
  }
  ret += tmp2;
  while (count > 0) {
    ret += *(a++) * *(b++);
    count--;
  }
#endif
#if 0
  while (count >= 4) {
    ret += _mm_cvtss_f32(_mm_dp_ps(_mm_set_ps(a[0],a[1],a[2],a[3]),
				   _mm_set_ps(b[0],b[1],b[2],b[3]), 0xff));
    a += 4;
    b += 4;
    count -= 4;
  }
  while (count > 0) {
    ret += *(a++) * *(b++);
    count--;
  }
#endif
#if 0
  size_t i;
  for (i = 0; i < count; i++) {
    ret += *(a++) * *(b++);
  }
#endif
#endif
  return ret;
}

// a[x] += b[x] * mul  {x=0..num}
void madd(float *a, const float *b, float m, int num) {
#ifdef __AVX__
  while (num > 0 && (((unsigned long long)a) & 31)) {
    *(a++) += *(b++) * m;
    num--;
  }
  __m256 tmp = _mm256_set_ps(m, m, m, m, m, m, m, m);
  while (num >= 8) {
    _mm256_store_ps(
      a,
      _mm256_add_ps(
	_mm256_load_ps(a),
	  _mm256_mul_ps(tmp, _mm256_loadu_ps(b))));
    a+=8;
    b+=8;
    num-=8;
  }
#endif
  while (num > 0) {
    *(a++) += *(b++) * m;
    num--;
  }
}

// x=0..num =
// sum( (a[x]*m) ^ 2)  =
// sum( a[x]^2 * m^2 )
// sum( a[x]^2 ) * m^2
float sqr_mul_sum(const float* a, float m, int num) {
  float ret = 0.0;
#ifdef __AVX__
  while (num > 0 && (((unsigned long long)a) & 31)) {
    ret += sqr(*(a++));
    num--;
  }
  __m256 sum = _mm256_set_ps(0.0, 0.0, 0.0, 0.0,
			     0.0, 0.0, 0.0, 0.0);
  while (num >= 8) {
    __m256 tmp = _mm256_load_ps(a);
    sum = _mm256_add_ps(sum, _mm256_mul_ps(tmp, tmp));
    a += 8;
    num -= 8;
  }
  ret += sum[0] + sum[1] + sum[2] + sum[3] + sum[4] + sum[5] + sum[6] + sum[7];
#endif  
  while (num > 0) {
    ret += sqr(*(a++));
    num--;
  }
  return ret * m * m;
}

// max( @abs(a[x]), cl) x = 0..num
float max_abs(const float* a, float cl, int num) {
#ifdef __AVX__
  while (num > 0 && (((unsigned long long)a) & 31)) {
    cl = MAX(ABS(*(a++)), cl);
    num--;
  }
  __m256 tmp = _mm256_set_ps(cl, cl, cl, cl, cl, cl, cl, cl);

  const __m256 kMask = _mm256_castsi256_ps(_mm256_set1_epi32(0x7fffffff));
  while (num >= 8) {
    tmp = _mm256_max_ps(tmp, _mm256_and_ps(kMask, _mm256_load_ps(a)));
    a += 8;
    num -= 8;
  }
  cl = MAX(MAX(MAX(tmp[0], tmp[1]), MAX(tmp[2], tmp[3])),
           MAX(MAX(tmp[4], tmp[5]), MAX(tmp[6], tmp[7])));
#endif
  while (num > 0) {
    cl = MAX(ABS(*(a++)), cl);
    num--;
  }
  return cl;
}

int h12(int mode,
	int lpivot,
	int l1,
	int m,
	float *u,
	float *up,
	float *cm,
	int icv,
	int ncv) {
  float d1,  b, clinv, cl, sm;
  int k, j;

  /* Check parameters */
  if (mode!=1 && mode!=2) 
    return(1);
  if (m<1 || u==NULL || cm==NULL) 
    assert(0);
  //     return(1);
  if (lpivot<0 || lpivot>=l1 || l1>m) 
    //     assert(0);
    return(1);

  /* Function Body */
  cl = ABS( u[lpivot] );
  // cl= (d1 = u[lpivot], fabs(d1));

  if (mode==2) 
  { /* Apply transformation I+U*(U**T)/B to cm[] */
    if(cl<=0.) 
      //       assert(0);
      return(0);
  } 
  else 
  {   /* Construct the transformation */


    /* This is the way provided in the original pseudocode
       sm = 0;
       for (j = l1; j < m; j++)
       {
       d1 =  u[j * u_dim1];
       sm += d1*d1;
       }
       d1 = u[lpivot * u_dim1];
       sm += d1*d1;
       sm = sqrt(sm);

       if (u[lpivot] > 0) 
       sm=-sm;

       up[0] = u[lpivot] - sm; 
       u[lpivot]=sm;
       printf("Got sum: %f\n",sm);
       */

    /* and this trying to compensate overflow */
#if 0
    for (j=l1; j<m; j++) 
    {  // Computing MAX 
      cl = MAX( ABS( u[j] ), cl );
    }
#else
    cl = max_abs(&u[l1], cl, m - l1);
#endif
    // zero vector?   

    if (cl<=0.) 
      return(0);

    clinv=1.0/cl;

    // Computing 2nd power 
    sm = sqr(u[lpivot] * clinv);

#if 0
    for (j=l1; j<m; j++) sm += sqr(u[j]*clinv);
#else
    sm += sqr_mul_sum(&u[l1], clinv, m - l1);
#endif    
    cl *= sqrt(sm); 

    if (u[lpivot] > 0.) 
      cl=-cl;
    up[0] = u[lpivot] - cl; 
    u[lpivot]=cl;
  }

  // no vectors where to apply? only change pivot vector! 
  b=up[0] * u[lpivot];

  /* b must be nonpositive here; if b>=0., then return */
  if (b >= 0) 
    return 0;

  // ok, for all vectors we want to apply
  for (j =0; j < ncv; j++) {
    sm = cm[ lpivot + j * icv ] * up[0];

#if 1
    sm += dot(&cm[l1 + j * icv], &u[l1], m - l1);
#else
    for (k=l1; k<m; k++) 
      sm += cm[ k + j*icv ] * u[ k ]; 
#endif

    if (sm != 0.0) {
      sm *= (1/b); 
      // cm[lpivot, j] = ..
      cm[ lpivot + j*icv] += sm * (up[0]);
#if 0
      for (k= l1; k<m; k++) 
      {
        cm[ k + j*icv] += u[k]*sm;
      }
#else
      madd(&cm[ l1 + j * icv], &u[l1], sm, m - l1);
#endif
    }
  }

  return(0);
} 


void g1(float a, float b, float *cterm, float *sterm, float *sig)
{
  float d1, xr, yr;

  if( fabs(a) > fabs(b) ) {
    xr = b / a;
    d1 = xr;
    yr = sqrt(d1*d1 + 1.);
    d1 = 1./yr;
    *cterm=(a>=0.0 ? fabs(d1) : -fabs(d1));
    *sterm=(*cterm)*xr;
    *sig=fabs(a)*yr;
  } else if( b != 0.) {
    xr = a / b;
    d1 = xr;
    yr = sqrt(d1 * d1 + 1.);
    d1 = 1. / yr;
    *sterm=(b>=0.0 ? fabs(d1) : -fabs(d1));
    *cterm=(*sterm)*xr; *sig=fabs(b)*yr;
  } else {
    *sig=0.0; *cterm=0.0; *sterm=1.0;
  }
} 


int nnls_algorithm(float *a, int m,int n, float *b, float *x, float *rnorm) {
  int pfeas;
  int ret=0;
  int iz;
  int jz;
  int k, j=0, l, itmax, izmax=0, ii, jj=0, ip;
  float d1, d2, sm, up, ss; 
  float temp, wmax, t, alpha, asave, dummy, unorm, ztest, cc;


  /* Check the parameters and data */
  if(m <= 0 || n <= 0 || a == NULL || b == NULL || x == NULL) 
    return(2);

  /* Allocate memory for working space, if required */
  float *w = (float*)calloc(n, sizeof(float));
  float *zz = (float*)calloc(m, sizeof(float));
  int *index = (int*)calloc(n, sizeof(int));
  if(w == NULL || zz == NULL || index == NULL) 
    return(2);

  /* Initialize the arrays INDEX[] and X[] */
  for(k=0; k<n; k++) {
    x[k]=0.0;
    index[k]=k;
  }

  int iz2 = n - 1;
  int iz1 = 0;
  int iter=0; 
  int nsetp=0;
  int npp1=0;

  /* Main loop; quit if all coeffs are already in the solution or */
  /* if M cols of A have been triangularized */
  if(n < 3) 
    itmax=n*3;
  else 
    itmax=n*n;
 

  while(iz1 <= iz2 && nsetp < m) {
    /* Compute components of the dual (negative gradient) vector W[] */
    for(iz=iz1; iz<=iz2; iz++) {
      j=index[iz];
#if 0
      sm=0.0;
      for(l=npp1; l<m; l++) sm+=a[j*m + l]*b[l];
      w[j]=sm;
#else
      w[j] = dot(&a[j*m+npp1], &b[npp1], m - npp1);
#endif
    }

    while(1) {
      /* Find largest positive W[j] */
      for(wmax=0., iz=iz1; iz<=iz2; iz++) {
        j=index[iz];
	if(w[j]>wmax) {
	  wmax=w[j];
	  izmax=iz;
	}
      }

      /* Terminate if wmax<=0.0; */
      /* it indicates satisfaction of the Kuhn-Tucker conditions */
      if(wmax<=0.0)
        break;

      iz=izmax; 
      j=index[iz];

      /* The sign of W[j] is ok for j to be moved to set P. */
      /* Begin the transformation and check new diagonal element to avoid */
      /* near linear dependence. */
      asave=a[j*m + npp1];
      h12(1, npp1, npp1+1, m, &a[j*m +0], &up, &dummy, 1, 0);
      unorm=0.0;
      if(nsetp!=0){
        for(l=0; l<nsetp; l++) {
          d1=a[j*m + l]; 
          unorm+=d1*d1;
        }
      }

      unorm=sqrt(unorm);
      d2=unorm+(d1=a[j*m + npp1], fabs(d1)) * 0.01;
      if((d2-unorm)>0.) {
        /* Col j is sufficiently independent. Copy B into ZZ, update ZZ */
        /* and solve for ztest ( = proposed new value for X[j] ) */
        for(l=0; l<m; l++) zz[l]=b[l];
        h12(2, npp1, npp1+1, m, &a[j*m + 0], &up, zz, 1, 1);
        ztest=zz[npp1]/a[j*m +npp1];
        /* See if ztest is positive */
        if(ztest>0.) break;
      }

      /* Reject j as a candidate to be moved from set Z to set P. Restore */
      /* A[npp1,j], set W[j]=0., and loop back to test dual coeffs again */
      a[j*m+ npp1]=asave; w[j]=0.0;
    } /* while(1) */

    if(wmax<=0.0)
      break;

    /* Index j=INDEX[iz] has been selected to be moved from set Z to set P. */
    /* Update B and indices, apply householder transformations to cols in */
    /* new set Z, zero subdiagonal elts in col j, set W[j]=0. */
    for(l=0; l<m; ++l) 
      b[l]=zz[l];

    index[iz]=index[iz1];
    index[iz1]=j;
    iz1++;
    npp1++;
    nsetp=npp1;

    if(iz1<=iz2) {
     for(jz=iz1; jz<=iz2; jz++) {
        jj=index[jz];
        h12(2, nsetp-1, npp1, m, &a[j*m +0], &up, &a[jj*m +0], m, 1);
      }
    }

    if(nsetp!=m) {
      for(l=npp1; l<m; l++) 
        a[j*m +l]=0.0;
    }

    w[j]=0.0;

    /* Solve the triangular system; store the solution temporarily in Z[] */
    for(l=0; l<nsetp; l++) {
      ip=nsetp-(l+1);
      if(l!=0) {
	for(ii=0; ii<=ip; ii++)
	  zz[ii]-=a[jj*m + ii]*zz[ip+1];
      }
      jj=index[ip];
      zz[ip]/=a[jj*m +ip];
    }

    /* Secondary loop begins here */
    while(++iter < itmax) {
      /* See if all new constrained coeffs are feasible; if not, compute alpha */
      for(alpha = 2.0, ip = 0; ip < nsetp; ip++) {
        l=index[ip];
        if(zz[ip]<=0.) {
          t = -x[l]/(zz[ip]-x[l]);
          if(alpha > t) {
            alpha = t; 
            jj = ip - 1;
          }
        }
      }

      /* If all new constrained coeffs are feasible then still alpha==2. */
      /* If so, then exit from the secondary loop to main loop */
      if(alpha==2.0) 
        break;

      /* Use alpha (0.<alpha<1.) to interpolate between old X and new ZZ */
      for(ip=0; ip<nsetp; ip++) {
        l = index[ip];
        x[l] += alpha*(zz[ip]-x[l]);
      }

      /* Modify A and B and the INDEX arrays to move coefficient i */
      /* from set P to set Z. */
      k=index[jj+1]; pfeas=1;
      do {
        x[k]=0.0;
        if(jj!=(nsetp-1)) {
          jj++;
          for(j=jj+1; j<nsetp; j++) {
            ii=index[j]; index[j-1]=ii;
            g1(a[ii*m + (j-1)], a[ii*m + j], &cc, &ss, &a[ii*m + j-1]);
            for(a[ii*m + j]=0., l=0; l<n; l++) if(l!=ii) {
              /* Apply procedure G2 (CC,SS,A(J-1,L),A(J,L)) */
              temp=a[l*m + j-1];
              a[l*m + j-1]=cc*temp+ss*a[l*m + j];
              a[l*m + j]=-ss*temp+cc*a[l*m + j];
            }
            /* Apply procedure G2 (CC,SS,B(J-1),B(J)) */
            temp=b[j-1];
	    b[j-1]=cc*temp+ss*b[j];
	    b[j]=-ss*temp+cc*b[j];
          }
        }
        npp1=nsetp-1; nsetp--; iz1--; index[iz1]=k;

        /* See if the remaining coeffs in set P are feasible; they should */
        /* be because of the way alpha was determined. If any are */
        /* infeasible it is due to round-off error. Any that are */
        /* nonpositive will be set to zero and moved from set P to set Z */
        for(jj=0, pfeas=1; jj<nsetp; jj++) {
          k=index[jj];
	  if(x[k]<=0.0) {pfeas=0; break;}
        }
      } while(pfeas==0);

      /* Copy B[] into zz[], then solve again and loop back */
      for(k=0; k<m; k++) 
        zz[k]=b[k];
      for(l=0; l<nsetp; l++) {
        ip=nsetp-(l+1);
        if(l!=0) {
	  for(ii=0; ii<=ip; ii++)
	    zz[ii]-=a[jj*m + ii]*zz[ip+1];
	}
        jj=index[ip]; zz[ip]/=a[jj*m + ip];
      }
    } /* end of secondary loop */

    if(iter>=itmax) {
      ret = 1;
      break;
    }

    for(ip=0; ip<nsetp; ip++) {
      k=index[ip];
      x[k]=zz[ip];
    }
  } /* end of main loop */

  /* Compute the norm of the final residual vector */
  sm=0.0;

  if (rnorm != NULL) {
    if (npp1<m) 
      for (k=npp1; k<m; k++) 
        sm+=(b[k] * b[k]);
    else 
      for (j=0; j<n; j++) 
        w[j]=0.0;
    *rnorm=sqrt(sm);
  } 

  /* Free working space, if it was allocated here */
  free(w);
  free(zz);
  free(index);
  return(ret);
}
/* nnls_ */


float *nnls(float *a_matrix, float *b_matrix, int height, int width) {

  float *solution = (float*)calloc(height, sizeof(float));

  if(solution == NULL) {
    fprintf(stderr, "could not allocate enough memory for nnls\n");
    exit(EXIT_FAILURE);
  }

  int ret = nnls_algorithm(a_matrix, width, height, b_matrix, solution, NULL);
  if(ret == 1) {
    printf("NNLS has reached the maximum iterations\n");
  } else if(ret == 2) {
    fprintf(stderr, "NNLS could not allocate enough memory\n");
    exit(EXIT_FAILURE);
  }

  return solution;
}
