#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include "timing.h"
#include "sync.h"
#include "lightbuffer.h"

// #define PULSE
#define ALSA

namespace ChaosDisplay {

Sync<TimeSync> timesync;

#ifdef PULSE

#define BUFSIZE 4096

#include <pulse/simple.h>
#include <pulse/error.h>

void PlayAudio(const char *device, char *filename, unsigned int rate) {
  FILE* f = fopen(filename, "r");
  if (!f) {
    fprintf(stderr, "Failed to open file %s\n", filename);
    exit(1);
  }
  /* The Sample format to use */
  pa_sample_spec ss;
  ss.format = PA_SAMPLE_S16LE;
  ss.channels = 2;
  ss.rate = rate;
  pa_simple *s = NULL;
  int ret = 1;
  int error;
  /* Create a new playback stream */
  if (!(s = pa_simple_new(NULL, device, PA_STREAM_PLAYBACK, NULL, "playback", &ss, NULL, NULL, &error))) {
    fprintf(stderr, __FILE__": pa_simple_new() failed: %s\n", pa_strerror(error));
    exit(1);
  }
  for (;;) {
    uint8_t buf[BUFSIZE];
    ssize_t r;
    pa_usec_t latency;
    if ((latency = pa_simple_get_latency(s, &error)) == (pa_usec_t) -1) {
      fprintf(stderr, __FILE__": pa_simple_get_latency() failed: %s\n", pa_strerror(error));
      exit(1);
    }
    TimeSync ts;
    ts.t = Time();
    ts.media_time = latency * 1E-6;
    timesync.Set(ts);
    
    /* Read some data ... */
    if ((r = fread(buf, 1, sizeof(buf), f)) <= 0) {
      if (r == 0) /* EOF */
	break;
      fprintf(stderr, __FILE__": read() failed: %s\n", strerror(errno));
      exit(1);
    }
    /* ... and play it */
    if (pa_simple_write(s, buf, (size_t) r, &error) < 0) {
      fprintf(stderr, __FILE__": pa_simple_write() failed: %s\n", pa_strerror(error));
      exit(1);
    }
  }
  fclose(f);
  /* Make sure that every single sample was played */
  if (pa_simple_drain(s, &error) < 0) {
    fprintf(stderr, __FILE__": pa_simple_drain() failed: %s\n", pa_strerror(error));
    exit(1);
  }
}

#endif

#ifdef ALSA
#include <alsa/asoundlib.h>
void PlayAudio(const char *device, char *filename, unsigned int rate) {
  int i;
  int err;
  snd_pcm_t *playback_handle;
  snd_pcm_hw_params_t *hw_params;
  
  if ((err = snd_pcm_open (&playback_handle, device, SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
    fprintf (stderr, "cannot open audio device %s (%s)\n", 
	     device,
	     snd_strerror (err));
    exit (1);
  }
     
  if ((err = snd_pcm_hw_params_malloc (&hw_params)) < 0) {
    fprintf (stderr, "cannot allocate hardware parameter structure (%s)\n",
	     snd_strerror (err));
    exit (1);
  }
   
  if ((err = snd_pcm_hw_params_any (playback_handle, hw_params)) < 0) {
    fprintf (stderr, "cannot initialize hardware parameter structure (%s)\n",
	     snd_strerror (err));
    exit (1);
  }

  if ((err = snd_pcm_hw_params_set_rate_resample(playback_handle, hw_params, 1))) {
    printf("Resampling setup failed for playback: %s\n", snd_strerror(err));
    exit (1);
  }

  if ((err = snd_pcm_hw_params_set_access (playback_handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
    fprintf (stderr, "cannot set access type (%s)\n",
	     snd_strerror (err));
    exit (1);
  }
  
  if ((err = snd_pcm_hw_params_set_format (playback_handle, hw_params, SND_PCM_FORMAT_S16_LE)) < 0) {
    fprintf (stderr, "cannot set sample format (%s)\n",
	     snd_strerror (err));
    exit (1);
  }
  
  if ((err = snd_pcm_hw_params_set_rate_near (playback_handle, hw_params, &rate, 0)) < 0) {
    fprintf (stderr, "cannot set sample rate (%s)\n",
	     snd_strerror (err));
    exit (1);
  }
  
  if ((err = snd_pcm_hw_params_set_channels (playback_handle, hw_params, 2)) < 0) {
    fprintf (stderr, "cannot set channel count (%s)\n",
	     snd_strerror (err));
    exit (1);
  }
  
  if ((err = snd_pcm_hw_params (playback_handle, hw_params)) < 0) {
    fprintf (stderr, "cannot set parameters (%s)\n",
	     snd_strerror (err));
    exit (1);
  }
  
  snd_pcm_hw_params_free (hw_params);
  
  if ((err = snd_pcm_prepare (playback_handle)) < 0) {
    fprintf (stderr, "cannot prepare audio interface for use (%s)\n",
	     snd_strerror (err));
    exit (1);
  }

  FILE* f = fopen(filename, "rb");
  if (!f) {
    perror("fopen");
    exit(1);
  }
  short buf[8192];
  int samples_played = 0;
  struct timeval start = Time();
  while (1) {
    int n  = fread(buf, 2, sizeof(buf)/sizeof(buf[0]), f);
    if (n == 0) break;
    if (n < 0) {
      perror("read");
      exit(1);
    }
    n/=2;
    TimeSync ts;
    ts.t = Time();
    snd_pcm_sframes_t delay;
    snd_pcm_delay(playback_handle, &delay);
    ts.media_time = ((double)(samples_played - delay)) / rate;
//    fprintf(stderr, "MT: %f  TT: %f  DELAY: %ld\n", ts.media_time, Time() - start, delay);
    timesync.Set(ts);
//    fprintf(stderr, "SYNC: %f\n", ts.media_time);
    if ((err = snd_pcm_writei (playback_handle, buf, n)) != n) {
      fprintf (stderr, "write to audio interface failed (%s)\n",
	       snd_strerror (err));
      exit (1);
    }
    samples_played += n;
  }
  
  snd_pcm_close (playback_handle);
}
#endif

void ShowFrame(FILE* f, LightBuffer *lb) {
  if (getc(f) != 0xEE) {
    fprintf(stderr, "Not at frame boundary.\n");
    exit(1);
  }
  // Frame
  for (int s = 0; ; s++) {
    int lights = getc(f);
    if (lights == EOF) {
      fprintf(stderr, "End of file!\n");
      exit(1);
    }
    lights <<= 8;
    lights |= (unsigned char)getc(f);
    if (lights == 0) break;
    for (int l = 0; l < lights; l++) {
      int r = getc(f);
      if (r == EOF) {
	fprintf(stderr, "End of file!\n");
	exit(1);
      }
      int g = getc(f);
      int b = getc(f);
      lb->Set(s, l, Color(r, g, b));
    }
  }
  lb->Show();
}

void PlayVideo(const char* filename, double frame_rate) {
  LightBuffer *lb = GetLightBuffer();
  FILE* f = fopen(filename, "rb");
  ShowFrame(f, lb);
  int frame_size = ftell(f);
  int last_frame = 0;
  while (1) {
    TimeSync ts = timesync.Get();
    struct timeval tv = Time();
    gettimeofday(&tv, NULL);
    double media_time = ts.media_time + (tv - ts.t);
    int frame = (int)(media_time * frame_rate);
    if (frame == last_frame) {
      double sleep = (frame + 1) / frame_rate - media_time;
      sleep = std::min(sleep, 1/frame_rate);
//      fprintf(stderr, "SLEEP: %f\n", sleep);
      Sleep(sleep);
      continue;
    }
    fprintf(stderr, "  FRAME = %d  media_time: %f  ", frame, media_time);
    fseek(f, frame * frame_size, SEEK_SET);
    ShowFrame(f, lb);
    last_frame = frame;
  }
  fclose(f);
}

void* StartVideo(void* arg) {
  char** argv = (char**) arg;
  PlayVideo(argv[3], strtod(argv[4], NULL));
}

};

int main(int argc, char** argv) {
  pthread_t video_thread;

  TimeSync ts;
  ts.t = Time();
  ts.media_time = 0.0;
  ChaosDisplay::timesync.Set(ts);
  
  pthread_create(&video_thread, NULL, &ChaosDisplay::StartVideo, argv);
  ChaosDisplay::PlayAudio("default", argv[1], atoi(argv[2]));
  pthread_join(video_thread, NULL);
  return 0;
}
