#include <vector>
#include <algorithm>
#include <stdio.h>
#include "lightbuffer.h"
#include "math.h"
#include "timing.h"
#include "stdlib.h"

struct FloatColor {
  FloatColor() : r(0), g(0), b(0) {}
  FloatColor(float r, float g, float b) : r(r), g(g), b(b) {}
  Color GetColor() {
    int red = r * 100;
    int green = g * 100;
    int blue = b * 100;
    if (red > 255) red = 255;
    if (green > 255) green = 255;
    if (blue > 255) blue = 255;
//    if (r != 0.0 || b != 0.0 || g != 0.0)
//      fprintf(stderr, "%f %f %f    => %d, %d, %d\n", r, g, b, red, green, blue);
    return Color(red, green, blue);
  }
  FloatColor operator*(float x) const { return FloatColor(r*x, g*x, b*x); }
  void operator+=(const FloatColor& c) { r+=c.r; g+=c.g; b+=c.b; }
  void Reset() { r=g=b=0.0; }

  float r, g, b;
};

float frand() {
  return (rand() & 0x7fffffff) / (float)0x7fffffff;
}


std::vector<FloatColor> output;

class Particle {
public:
  Particle() {
    pos = frand();
    fprintf(stderr, "POS= %f\n", pos);
    speed = frand() * 2 - 1;
    color.r = frand();
    color.g = frand();
    color.b = frand();
  }
  void Draw() {
    int p = (int)(pos * output.size());
    for (int i = p - 5; i < p + 5; i++) {
      float dist = (6 - fabs(pos - i / (float)output.size())) / 6;
      output[(i+output.size()) % output.size()] += color * dist;
    }
//    output[p]+=FloatColor(3,3,3);
//    fprintf(stderr, "DRAW, p=%d output[p]=%f, %f, %f\n", p, output[p].r, output[p].g, output[p].b);
  }

  void Update(float mult) {
#if 1
    if (boom) {
      fprintf(stderr, "BOOM!!!!\n");
      speed = frand() * 6 - 3;
      boom = false;
    }
#endif
//    fprintf(stderr, "Gravity = %e\n", gravity);
    speed += gravity * mult / 200;
    pos += speed * mult;
    speed *= 0.99;
  }

  float operator-(const Particle& other) const {
    float dist = pos - other.pos;
    if (dist < -0.5) dist += 1.0;
    if (dist > 0.5) dist -= 1.0;
    return dist;
  }

  bool operator<(const Particle& other) const {
    return pos < other.pos;
  }

  bool boom;
  int close;
  float gravity;
  float pos;    // [0..1)
  float speed;  //
  FloatColor color;
};

class System {
public:
  System(int howmany) : particles(howmany) {
    for (int i = 0; i < howmany; i++) {
      particles[i] = Particle();
    }
  }
  
  void calculate_gravity() {
    std::sort(particles.begin(), particles.end());
    for (int i = 0; i < particles.size(); i++) {
      particles[i].gravity = 0.0;
      particles[i].close = 0;
    }
    for (int i = 0; i < particles.size(); i++) {
      for (int j = i + 1; i < particles.size(); i++) {
	float dist = (particles[i] - particles[j]);
	float absdist = fabs(dist);
	if (absdist < 1.0/1000) {
	  particles[i].close++;
	  particles[j].close++;
	  continue;
	}
	if (absdist < 1.0/10) {
	  absdist = 1.0/10;
	}
	// mass = particles[i].mass * particles[j].mass; // assumed 1.0
	float d2 = 1.0 / (dist * absdist);
//	fprintf(stderr, "d2 = %e  dist = %e  absdist = %e\n", d2, dist, absdist);
//	exit(1);
	particles[i].gravity -= d2;
	particles[j].gravity += d2;
      }
    }
    for (int i = 0; i < particles.size(); i++) {
      if (particles[i].close > 10) {
	for (int j = 0; j < particles.size(); j++) {
	  if (fabs(particles[j] - particles[i]) < 2.0/1000) {
	    particles[j].boom = true;
	  }
	}
      }
    }
  }

  void update(float mult) {
    calculate_gravity();
    for (int i = 0; i < particles.size(); i++) particles[i].Update(mult);
    for (int i = 0; i < output.size(); i++) output[i].Reset();
    for (int i = 0; i < particles.size(); i++) particles[i].Draw();
  }

  std::vector<Particle> particles;
};

int main(int argc, char **argv) {
  LightBuffer* l = GetLightBuffer();
  std::vector<int> geometry;
  l->GetGeometry(&geometry);
  int total = 0;
  for (int i = 0; i < geometry.size(); i++)
    total += geometry[i];
  output.resize(total);
  System system(300);

  l->Show();
  l->Show();
  l->Show();
  double speed = 1.0/30;
  int frames = 0;
  struct timeval start = Time();
  while (1) {
    system.update(speed / 10);
    int n = 0;
    for (int i = 0; i < geometry.size(); i++) {
      for (int lamp = 0; lamp < geometry[i]; lamp++) {
	l->Set(i, lamp, output[n++].GetColor());
      }
    }
    l->Show();
    frames++;
    speed = (Time() - start) / frames;
  }
  
}
