#include <string>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <memory.h>
#include "matrix.h"
#include "lightbuffer.h"
#include "images.h"
#include <map>
#include <math.h>
#include <deque>
#include <algorithm>

// #define USE_SOLUTION
// #define DUMP

#define XSIZE 231
#define YSIZE 178

#define STR2(X) #X
#define STR(X) STR2(X)

void DumpMat(Matrix<float> m, const std::string filename) {
  FILE *f = fopen(filename.c_str(), "w");
  fprintf(f, "# Matrix\n");
  fprintf(f, "# rows: %d\n", m.rows());
  fprintf(f, "# columns: %d\n", m.columns());
  for (int y = 0; y < m.rows(); y++) {
    for (int x = 0; x < m.columns(); x++) {
      fprintf(f, "%f ", m.get(x, y));
    }
    fprintf(f, "\n");
  }
  fclose(f);
}

void DumpSparse(Matrix<float> m, const std::string filename) {
  FILE *f = fopen(filename.c_str(), "w");
  int values = 0;
  for (int y = 0; y < m.rows(); y++)
    for (int x = 0; x < m.columns(); x++)
      if (m.get(x, y) != 0.0)
	values++;
  fprintf(f, "SPARSE\n%d %d\n%d\n", m.rows(), m.columns(), values);
  for (int y = 0; y < m.rows(); y++)
    for (int x = 0; x < m.columns(); x++)
      if (m.get(x, y) != 0.0)
	fprintf(f, "%d %d %10.16g\n", y + 1, x + 1, m.get(x, y));
  fclose(f);
}

const double squish_factor = 1.2375;

class YuvStream {
public:
  YuvStream(const std::string &filename) {
    cdiv = 2;
#if 0
    setenv("FILE", filename.c_str(), 1);
    system("mkfifo -m 660 /tmp/stream.yuv");
    system("mkfifo -m 660 /tmp/follow.yuv");
    system("mplayer -ao null -vo yuv4mpeg:file=/tmp/stream.yuv -zoom -vf scale -xy 160 -sws 9 \"$FILE\" &");
    system("mplayer /tmp/follow.yuv &");
    f = fopen("/tmp/stream.yuv", "r");
    out = fopen("/tmp/follow.yuv", "w");
    eof_reached = 0;
#else
    setenv("FILE", filename.c_str(), 1);
    int xsize;
    int ysize;
    f = popen("ffmpeg -i \"$FILE\" 2>&1 | sed -n 's@.*DAR \\([0-9]*\\):\\([0-9]*\\)[^0-9].*@\\1 \\2@gp'", "r");
    if (fscanf(f, "%d %d", &xsize, &ysize) != 2) {
      fprintf(stderr, "Failed to get aspect!\n");
      exit(1);
    }
    fclose(f);
    fprintf(stderr, "Input aspect %d / %d\n", xsize, ysize);
    
    double yscale = (double)YSIZE / (double)ysize;
    double xscale = yscale / squish_factor;
#if 0
    // No squish for 4:3 videos
    if (xsize * 3 < 4 * ysize) {
      xscale = yscale;
    }
#endif
    int tmp_xsize = (int)(xsize * xscale);
    char strtmp[256];
    sprintf(strtmp, "scale=%d:%d", tmp_xsize, YSIZE);
    std::string filter(strtmp);
    if (tmp_xsize > XSIZE) {
      sprintf(strtmp, ",crop=%d:%d", XSIZE, YSIZE);
      filter += std::string(strtmp);
    } else if (tmp_xsize < XSIZE) {
      sprintf(strtmp, ",pad=%d:%d:(ow-iw)/2:(oh-ih)/2", XSIZE, YSIZE);
      filter += std::string(strtmp);
    }
    
    std::string cmd = "ffmpeg -i \"$FILE\" -an -vf '" + filter + "' -f yuv4mpegpipe -";
    //    f = popen("avconv -i \"$FILE\" -s 160x90 -an -f yuv4mpegpipe -", "r");
    // f = popen("avconv -i \"$FILE\" -s " STR(XSIZE) "x" STR(YSIZE) " -an -f yuv4mpegpipe -", "r");
    fprintf(stderr, "Executing: %s\n", cmd.c_str());
    f = popen(cmd.c_str(), "r");
    out = NULL;
    eof_reached = 0;
#endif
    read_header();
  }

  int c() {
    int c = getc(f);
    if (out) putc(c, out);
    return c;
  }

  int getnum() {
    int x = 0;
    while (1) {
      int y = c();
      y -= '0';
      if (y < 0 || y > 9) return x;
      x *= 10;
      x += y;
    }
  }

  void read_header() {
    int k;
    if((k=c()) != 'Y' ||
       (k=c()) != 'U' ||
       (k=c()) != 'V' ||
       (k=c()) != '4' ||
       (k=c()) != 'M' ||
       (k=c()) != 'P' ||
       (k=c()) != 'E' ||
       (k=c()) != 'G' ||
       (k=c()) != '2')
      goto headerr;

    while(1)
    {
      switch(c())
      {
	case ' ':
	  continue;
	case 'W':
	  xsize = getnum();
	  break;
	case 'H':
	  ysize = getnum();
	  break;
	case 'C':
	  cformat = getnum();
	  switch (cformat) {
	    case 444:
	      cdiv = 1;
	      break;
	  }
	  break;

	case '\n': return;
	case EOF: goto headeof;
	default:
	  while(1) {
	    switch(c()) {
	      case '\n':
		return;
	      case ' ':
		break;
	      case EOF:
		goto headeof;
	      default:
		continue;
	    }
	    break;
	  }
      }
    }

  headerr:
    perror("**while reading header");
    exit(1);

  headeof:
    fprintf(stderr,"**End of file in header\n");
    exit(1);
  }

  static float gamma(float x) {
    if (x < 0.0) x = 0.0;
    if (x > 1.0) x = 1.0;
    x = pow(x, 2.2);
    return x;
  }


  bool read_frame(Matrix<float> *r, Matrix<float> *g, Matrix<float> *b)
  {
    int uv_x = xsize / cdiv;
    int uv_y = ysize / cdiv;
    Matrix<float> y_comp(xsize, ysize);
    Matrix<float> u_comp(uv_x, uv_y);
    Matrix<float> v_comp(uv_x, uv_y);
    
    int k;
    struct frame *ret;
    if(eof_reached) {
      fprintf(stderr, "AT EOF!\n");
      return false;
    }
    switch ((k=c())) {
      case 'F': break;
      case EOF: goto stream_end;
      default: goto frameerr;
    }
    if((k=c()) != 'R' ||
       (k=c()) != 'A' ||
       (k=c()) != 'M' ||
       (k=c()) != 'E')
      goto frameerr;

    while (1) {
      switch(k=c()) {
	case '\n': break;
	case EOF: goto frameeof;
	default: continue;
      }
      break;
    }
    for (int y = 0; y < ysize; y++) {
      for (int x = 0; x < xsize; x++) {
	y_comp.set(x, y, c());
      }
    }
    for (int y = 0; y < uv_y; y++) {
      for (int x = 0; x < uv_x; x++) {
	u_comp.set(x, y, c());
      }
    }
    for (int y = 0; y < uv_y; y++) {
      for (int x = 0; x < uv_x; x++) {
	v_comp.set(x, y, c());
      }
    }
#if 0
    {
      float Y = y_comp.get(0,0);
      float U = u_comp.get(0,0);
      float V = v_comp.get(0,0);
    fprintf(stderr, "YUV<%f %f %f> = SRGB<%f %f %f> = RGB<%f %f %f>\n",
	    y_comp.get(0,0),
	    u_comp.get(0,0),
	    v_comp.get(0,0),
	    ((1.164*(Y - 16) + 1.596*(V - 128))) / 256,
	    ((1.164*(Y - 16) - 0.813*(V - 128) - 0.391*(U - 128))) / 256,
	    ((1.164*(Y - 16) + 2.018*(U - 128))) / 256,
	    pow( ((1.164*(Y - 16) + 1.596*(V - 128))) / 256, 2.2),
	    pow( ((1.164*(Y - 16) - 0.813*(V - 128) - 0.391*(U - 128))) / 256, 2.2),
	    pow( ((1.164*(Y - 16) + 2.018*(U - 128))) / 256, 2.2));
    }
#endif
    
    for (int y = 0; y < ysize; y++) {
      for (int x = 0; x < xsize; x++) {
	float Y = y_comp.get(x, y);
	int ux = x/cdiv;
	int uy = y/cdiv;
	if (ux == u_comp.columns()) ux--;
	if (uy == u_comp.rows()) uy--;
	float U = u_comp.get(ux, uy);
	float V = v_comp.get(ux, uy);
	float R = (1.164*(Y - 16) + 1.596*(V - 128)) / 256;
	float G = (1.164*(Y - 16) - 0.813*(V - 128) - 0.391*(U - 128)) / 256;
	float B = (1.164*(Y - 16) + 2.018*(U - 128)) / 256;

#if 1
	R = gamma(R);
	G = gamma(G);
	B = gamma(B);
#endif

	r->set(x, y, R);
	g->set(x, y, G);
	b->set(x, y, B);
      }
    }
#if 0
    fprintf(stderr, "RGB<%f %f %f>\n",
	    r->get(0,0),
	    g->get(0,0),
	    b->get(0,0));
#endif
    return true;

  frameeof:
    fprintf(stderr,"**EOF while reading frame\n");
    return false;

  stream_end:
    fprintf(stderr,"**EOF!\n");
    return false;

  frameerr:
    fprintf(stderr,"**Unexpected data while reading frame: %c (%d)\n",k,k);
    for (int i = 0; i < 100; i++) {
      fprintf(stderr, "k = %c ", c());
    }
    return false;
  }

private:
  FILE *f;
  FILE *out;
  int xsize, ysize;
  int cformat;
  int cdiv;
  int eof_reached;
};

int scalecolor(float n) {
  int x = float(n * 256);
  if (x > 255) x = 255;
  if (x < 0) x = 0;
  return x;
}

Matrix<float> read_solution(int lights, int strings) {
  Matrix<float> solution = read_matrix("solution");
  // Normalize the solution so that an input image
  // of all ones outputs just enough energy to light
  // up all the lights.
#if 0
  Matrix<float> tmp(XSIZE, YSIZE);
  tmp = 1.0;
  tmp.Reshape(1, tmp.columns() * tmp.rows());
#else
  Matrix<float> tmp(1, XSIZE * YSIZE);
  tmp += 1.0;
#endif
  tmp = solution * tmp;
  double sum = 0.0;
  for (int l = 0; l < lights; l++)
    for (int s = 0; s < strings; s++)
      sum += tmp.get(0, l + s * lights);
  return solution * (lights * strings / sum);
}

Matrix<float> read_images(int lights) {
  Matrix <float> tmp;
  for (int l = 0; l < lights; l++) {
    char filename[512];
    sprintf(filename,"light-matrices/%06d.matrix", l);
    Matrix<float> lm = read_matrix(filename);
    if (tmp.columns() == 0) {
      tmp = Matrix<float>(lights, lm.columns() * lm.rows());
    }
    float mult = 1.0;
    float threshold = 0.045;
#if 0
    float energy = 0.0;
    float max = 0.0;
    // per-light normalization
    for (int y = 0; y < lm.rows(); y++) {
      for (int x = 0; x < lm.columns(); x++) {
	energy += lm.get(x, y);
	max = std::max(max, lm.get(x, y));
      }
    }
    if (energy != 0.0) {
      threshold = max / 2;
//	mult = 1.0/energy;
    } else {
      fprintf(stderr, "ZERO ENERGY\n");
    }
#endif
    for (int y = 0; y < lm.rows(); y++) {
      for (int x = 0; x < lm.columns(); x++) {
	float v = lm.get(x, y) * mult;
	if (v < threshold) v = 0.0;
	tmp.set(l, y * lm.columns() + x, v);
      }
    }
  }
  
  return tmp;
}

class SolverBase {
public:
  virtual ~SolverBase() {}
  virtual Matrix<float> Solve(Matrix<float> b) = 0;
};

class UseSolutionSolver : public SolverBase {
public:
  UseSolutionSolver(int lights, int strings) {
    solution_ = read_solution(lights, strings);
  }
  virtual Matrix<float> Solve(Matrix<float> b) {
    return solution_ * b;
  }
private:
  Matrix<float> solution_;
};

class NNLSSolverSimple : public SolverBase {
public:
  NNLSSolverSimple(int lights) {
    solution_ = read_images(lights);
  }
  virtual Matrix<float> Solve(Matrix<float> b) {
    return solve_nnls(solution_, b);
    // return solution_.fnnls(b);
  }
private:
  Matrix<float> solution_;
};

class SplitNNLSSolver : public SolverBase {
public:
  struct SubProblem {
    std::vector<int> rows;
    std::vector<int> columns;
    Matrix<float> slice;
  };

  SplitNNLSSolver(int lights)
    : lights_(lights) {
    Matrix<float> tmp = read_images(lights);

    std::vector<int> row_subproblems(tmp.rows(), -1);
    std::vector<int> col_subproblems(tmp.columns(), -1);
    int num_subproblems = 0;
    int unused_lights = 0;
    for (int light = 0; light < tmp.columns(); light++) {
      if (col_subproblems[light] != -1) continue;
      struct SubProblem subproblem;
      col_subproblems[light] = num_subproblems;
      std::deque<std::pair<int, int> > to_process;
      to_process.push_back(std::make_pair(0, light));
      subproblem.columns.push_back(light);
      while (!to_process.empty()) {
	if (to_process.front().first == 0) {
	  int x = to_process.front().second;
	  for (int y = 0; y < tmp.rows(); y++) {
	    if (tmp.get(x, y) == 0.0) continue;
	    if (row_subproblems[y] == -1) {
	      row_subproblems[y] = num_subproblems;
	      to_process.push_back(std::make_pair(1, y));
	      subproblem.rows.push_back(y);
	    }
	  }
	} else {
	  int y = to_process.front().second;
	  for (int x = 0; x < tmp.columns(); x++) {
	    if (tmp.get(x, y) == 0.0) continue;
	    if (col_subproblems[x] == -1) {
	      col_subproblems[x] = num_subproblems;
	      to_process.push_back(std::make_pair(0, x));
	      subproblem.columns.push_back(x);
	    }
	  }
	}
	to_process.pop_front();
      }
      if (subproblem.rows.empty()) {
	unused_lights++;
	continue;
      }

      subproblem.slice = Matrix<float>(subproblem.columns.size(),
				       subproblem.rows.size());
      for (int y = 0; y < subproblem.rows.size(); y++) {
	for (int x = 0; x < subproblem.columns.size(); x++) {
	  subproblem.slice.set(x, y, tmp.get(subproblem.columns[x],
					     subproblem.rows[y]));
	}
      }
      subproblems_.push_back(subproblem);
      num_subproblems++;
    }

    fprintf(stderr, "Num subproblems: %d Unused lights: %d\n", num_subproblems, unused_lights);
  }

  virtual Matrix<float> Solve(Matrix<float> b) {
    Matrix<float> ret(1, lights_);
    for (int p = 0; p < subproblems_.size(); p++) {
      Matrix<float> tmp(1, subproblems_[p].rows.size());
      for (int y = 0; y < subproblems_[p].rows.size(); y++) {
	tmp.set(0, y, b.get(0, subproblems_[p].rows[y]));
      }
      tmp = solve_nnls(subproblems_[p].slice, tmp);
      for (int x = 0; x < subproblems_[p].columns.size(); x++) {
	ret.set(0, subproblems_[p].columns[x], tmp.get(0, x));
      }
    }
    return ret;
  }

private:
  int lights_;
  std::vector<SubProblem> subproblems_;
};

class LightFinderThreader : public Threader {
public:
  LightFinderThreader(int lights,
		      std::vector<float> &x,
		      std::vector<float> &y,
		      std::vector<float> &mv,
		      std::vector<float> &bad) : x_(x), y_(y), mv_(mv), bad_(bad) {
    Run(lights);
  }
  std::vector<float> &x_;
  std::vector<float> &y_;
  std::vector<float> &mv_;
  std::vector<float> &bad_;

  void DoTask(int l) override {
    char filename[512];
    sprintf(filename,"light-matrices/%06d.matrix", l);
    fprintf(stderr, "Reading %d\n", l);
    Matrix<float> image = read_matrix(filename);
    float xw = 0.0;
    float yw = 0.0;
    float total = 0.0;
    float max_value = 0.0;
    int xp = 0;
    int yp = 0;
    for (int y = 0; y < image.rows(); y++) {
      for (int x = 0; x < image.columns(); x++) {
	float value = image.get(x, y);
	if (value > max_value) {
	  max_value = value;
	  xp = x;
	  yp = y;
	}
      }
    }
//      fprintf(stderr, "Max = %f\n", max_value);
    for (int y = 0; y < image.rows(); y++) {
      if (std::abs(yp - y) > 10) continue;
      for (int x = 0; x < image.columns(); x++) {
	if (std::abs(xp - x) > 10) continue;
	float value = image.get(x, y);
	value *= value;
	total += value;
	xw += x * value;
	yw += y * value;
      }
    }
    if (total != 0.0) {
      xw /= total;
      yw /= total;
    }
//      fprintf(stderr, "P = %f, %f\n", xw, yw);
    x_[l] = xw;
    y_[l] = yw;
    mv_[l] = max_value;
    bad_[l] = false;
  }
};

class BilinearSolver : public SolverBase {
public:
  BilinearSolver(int lights, const std::vector<int>& geometry) :
    lights_(lights), x_(lights), y_(lights), mv_(lights), bad_(lights) {
    int bad = 0;
    float lx = 0.0;
    float ly = 0.0;
#if 1
    LightFinderThreader findid(lights, x_, y_, mv_, bad_);
#else    
    for (int l = 0; l < lights; l++) {
      char filename[512];
      sprintf(filename,"light-matrices/%06d.matrix", l);
      fprintf(stderr, "Reading %d\n", l);
      Matrix<float> image = read_matrix(filename);
      float xw = 0.0;
      float yw = 0.0;
      float total = 0.0;
      float max_value = 0.0;
      int xp = 0;
      int yp = 0;
      for (int y = 0; y < image.rows(); y++) {
	for (int x = 0; x < image.columns(); x++) {
	  float value = image.get(x, y);
	  if (value > max_value) {
	    max_value = value;
	    xp = x;
	    yp = y;
	  }
	}
      }
//      fprintf(stderr, "Max = %f\n", max_value);
      for (int y = 0; y < image.rows(); y++) {
	if (std::abs(yp - y) > 10) continue;
	for (int x = 0; x < image.columns(); x++) {
	  if (std::abs(xp - x) > 10) continue;
	  float value = image.get(x, y);
	  value *= value;
	  total += value;
	  xw += x * value;
	  yw += y * value;
	}
      }
      if (total != 0.0) {
	xw /= total;
	yw /= total;
      }
//      fprintf(stderr, "P = %f, %f\n", xw, yw);
      x_[l] = xw;
      y_[l] = yw;
      mv_[l] = max_value;
      bad_[l] = false;
    }
#endif
    
    std::vector<float> dists(x_.size());
    for (int l = 0; l < lights - 1; l++) {
      dists[l] = dist(l, l + 1);
    }
    int l = 0;
    for (int i = 0; i < geometry.size(); i++) {
      l += geometry[i];
      dists[l - 1] = 0.0;
    }

    std::vector<float> tmp(dists);
    std::sort(tmp.begin(), tmp.end());
    median_dist_ = tmp[tmp.size() / 2];

    tmp = mv_;
    std::sort(tmp.begin(), tmp.end());
    median_value_ = tmp[tmp.size() / 2];

    std::vector<float> scores(lights);
    bool done = false;
    while (!done) {
      done = true;
      for (int l = 0; l < lights; l++) {
	float s = score(l);
	if (!bad_[l] && s > 50 && s > score(l-1) && s > score(l+1)) {
	  bad_[l] = true;
	  done = false;
	}
      }
    }

    int interpolated = 0;
    l = 0;
    for (int i = 0; i < geometry.size(); i++) {
      int last_good = lights;
      for (int n = 0; n < geometry[i]; n++, l++) {
	fprintf(stderr, "%d: %f %f (%f)   %f   %f   = %f  %s\n", l, x_[l], y_[l], mv_[l] / median_value_, dist(l, l-1) / median_dist_, dist(l, l+1) / median_dist_, score(l), bad_[l] ? "BAD" : "GOOD");
	if (bad_[l]) {
	  bad++;
	} else {
	  int delta = l - last_good;
	  if (delta > 1) {
	    for (int i = last_good + 1; i < l; i++) {
	      interpolated++;
	      x_[i] = (x_[last_good] * (l-i) + x_[l] * (i-last_good)) / delta;
	      y_[i] = (y_[last_good] * (l-i) + y_[l] * (i-last_good)) / delta;
	    }
	  }
	  last_good = l;
	}
      }
      fprintf(stderr, "%d bad lights, %d lights interpolated\n", bad, interpolated);
    }
  }

  float lookup(Matrix<float>& i, int x, int y) {
    if (x < 0) return 0.0;
    if (y < 0) return 0.0;
    if (x >= i.columns()) return 0.0;
    if (y >= i.rows()) return 0.0;
    return i.get(x, y);
  }

  virtual Matrix<float> Solve(Matrix<float> b) {
    Matrix<float> ret(1, lights_);
    b.Reshape(XSIZE, YSIZE);
    for (int l = 0; l < lights_; l++) {
      float x = x_[l];
      float y = y_[l];
      int xpos = floor(x);
      int ypos = floor(y);
      float xw2 = x - xpos;
      float xw1 = 1 - xw2;
      float yw2 = y - ypos;
      float yw1 = 1 - yw2;

      // Bilinear lookup
      ret.set(0, l,
	      (lookup(b, xpos, ypos) * xw1 + lookup(b, xpos + 1, ypos) * xw2) * yw1 +
	      (lookup(b, xpos, ypos+1) * xw1 + lookup(b, xpos + 1, ypos+1) * xw2) * yw2);
    }
    return ret;
  }

  float score(int a) {
    if (a < 0 || a >= lights_) return 0.0;
    if (bad_[a]) return 0.0;
    float d1 = dist(a, a-1);
    float d2 = dist(a, a+1);
    return (d1 + d2) / median_dist_ / (mv_[a] == 0.0 ? 0.1 : mv_[a] / median_value_);
  }

  float dist(int a, int b) {
    if (a < 0) return 0.0;
    if (b < 0) return 0.0;
    if (a >= x_.size()) return 0.0;
    if (b >= x_.size()) return 0.0;
    if (bad_[a] || bad_[b]) return 0.0;
    float xd = x_[a] - x_[b];
    float yd = y_[a] - y_[b];
    return sqrt(xd * xd + yd * yd);
  }

private:
  std::vector<float> x_;
  std::vector<float> y_;
  std::vector<float> mv_;
  std::vector<float> bad_;
  float median_dist_;
  float median_value_;
  int lights_;
};

class VideoThreader : public ThreaderBase {
public:
  VideoThreader(YuvStream *stream, LightBuffer *lb) :
    stream_(stream),
    lb_(lb),
    frame_num_(0),
    output_frame_num_(0) {
    std::vector<int> geometry;
    lb->GetGeometry(&geometry);
    lights_ = 0;
    for (int i = 0; i < geometry.size(); i++) {
      lights_ += geometry[i];
    }
    fprintf(stderr, "Generating video for %d lights\n", lights_);
#ifdef USE_SOLUTION
    solver_ = new UseSolutionSolver(lights_);
#else
#if 0
    // There is currently no normalization here...
    solver_ = new SplitNNLSSolver(lights_);
#endif
#if 1
    solver_ = new BilinearSolver(lights_, geometry);
#endif
#ifdef DUMP
    DumpSparse(solution, "A.sparse");
#endif
#endif

    Run();
  }
  struct SolutionHolder {
    Matrix<float> red_lights;
    Matrix<float> green_lights;
    Matrix<float> blue_lights;
  };
  virtual void Worker() {
    Matrix<float> r(XSIZE, YSIZE);
    Matrix<float> g(XSIZE, YSIZE);
    Matrix<float> b(XSIZE, YSIZE);
    mutex_.Lock();
    std::vector<int> geometry;
    lb_->GetGeometry(&geometry);
    while (true) {
      bool got_frame = stream_->read_frame(&r, &g, &b);
      int frame_num = frame_num_;
      frame_num_++;
//      if (frame_num > 30) got_frame = false;
      mutex_.Unlock();
      if (!got_frame) return;
      r.Reshape(1, r.columns() * r.rows());
      g.Reshape(1, g.columns() * g.rows());
      b.Reshape(1, b.columns() * b.rows());
      SolutionHolder out;
#ifdef DUMP
      char filename[256];
      sprintf(filename, "%06d_r.mat", frame_num);
      DumpMat(r, std::string(filename));
      sprintf(filename, "%06d_g.mat", frame_num);
      DumpMat(g, std::string(filename));
      sprintf(filename, "%06d_b.mat", frame_num);
      DumpMat(b, std::string(filename));
#endif

      out.red_lights = solver_->Solve(r);
      out.green_lights = solver_->Solve(g);
      out.blue_lights = solver_->Solve(b);

      r.Reshape(XSIZE, YSIZE);
      g.Reshape(XSIZE, YSIZE);
      b.Reshape(XSIZE, YSIZE);
      mutex_.Lock();
      output_queue_[frame_num] = out;
      while (output_queue_.begin() != output_queue_.end() &&
	     output_queue_.begin()->first == output_frame_num_) {
	out = output_queue_.begin()->second;
	output_queue_.erase(output_frame_num_);
	output_frame_num_++;

	int light = 0;
	for (int s = 0; s < geometry.size(); s++) {
	  for (int l = 0; l < geometry[s]; l++) {
	    lb_->Set(s, l, Color(scalecolor(out.red_lights.get(0, light)),
				 scalecolor(out.green_lights.get(0, light)),
				 scalecolor(out.blue_lights.get(0, light))));
	    light++;
	  }
	}
	lb_->Show();
      }
    }
  }

private:
  int lights_;
  YuvStream *stream_;
  LightBuffer *lb_;
  int frame_num_;
  int output_frame_num_;
  SolverBase* solver_;   // FIXME: delete this
  std::map<int, SolutionHolder> output_queue_;
};

int main(int argc, char **argv) {
  YuvStream stream(argv[1]);

  VideoThreader t(&stream, GetLightBufferStdout());
}
