#include <string>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <memory.h>
#include "matrix.h"
#include "lights.h"
#include "images.h"

class TestStream {
public:
  TestStream() {
    frame = 0;
  }

  void read_frame(Matrix<float> *r, Matrix<float> *g, Matrix<float> *b)
  {
    for (int y = 0; y < 90; y++) {
      for (int x = 0; x < 120; x++) {
	r->set(x, y, 0.0);
	g->set(x, y, 0.0);
	b->set(x, y, 0.0);
      }
    }

    int xpos = frame % 120;
    int ypos = (frame/2) % 90;
    int radius = 7;
    fprintf(stderr, "\rframe = %d xpos = %3d ypos = %3d", frame++, xpos, ypos);
    for (int y = ypos - radius; y <= ypos + radius; y++) {
      if (y < 0) continue;
      if (y >= 90) continue;
      for (int x = xpos - radius; x <= xpos + radius; x++) {
	if (x < 0) continue;
	if (x >= 120) continue;
	r->set(x, y, 1.0);
	g->set(x, y, 1.0);
	b->set(x, y, 1.0);
      }
    }
  }

  private:
  int frame;
};

int scalecolor(float n) {
  int x = float(n * 256);
  if (x > 255) x = 255;
  if (x < 0) x = 0;
  return x;
}

void test() {
  Matrix<float> a(3, 4);
  Matrix<float> b(1, 4);
  a.set(0, 0, 2.0);
  a.set(1, 0, -1.0);
  a.set(2, 0, 1.0);
  a.set(0, 1, 1.0);
  a.set(1, 1, -5.0);
  a.set(2, 1, 2.0);
  a.set(0, 2, -3.0);
  a.set(1, 2, 1.0);
  a.set(2, 2, -4.0);
  a.set(0, 3, 1.0);
  a.set(1, 3, -1.0);
  a.set(2, 3, 1.0);
  b.set(0, 0, -4.0);
  b.set(0, 1, 2.0);
  b.set(0, 2, 5.0);
  b.set(0, 3, -1.0);
  Matrix<float> answer = a.solve_least_square(b);
  fprintf(stderr, "[ %f,  %f,  %f ]\n", answer.get(0, 0), answer.get(0, 1), answer.get(0, 2));
  fprintf(stderr, "[ %f,  %f,  %f ]\n", -18.0/7.0, -151.0/210.0, 107.0/210.0);
}

void read_image(int s, int l, int x, Matrix<float> *channels) {
  char tmp[256];
  sprintf(tmp, "djpeg -scale 1/8 -pnm images/light_%02d_%02d_%d.jpg",
	  s, l, x);
  fprintf(stderr, "Running: %s\n", tmp);
  FILE *f = popen(tmp, "r");
  readpnm(f, channels);
  fclose(f);
}

Matrix<float> read_diff(int s, int l) {
  Matrix<float> images[8][3];
  for (int image = 0; image < 8; image++) {
    read_image(s, l, image, images[image]);
  }
  int xsize = images[0][0].columns();
  int ysize = images[0][0].rows();
  Matrix<float> ret(xsize, ysize);
  Matrix<float> eq(4, 8 * 3);
  Matrix<float> ans(1, 8 * 3);
  for (int y = 0; y < ysize; y++) {
    for (int x = 0; x < xsize; x++) {
      for (int image = 0; image < 8; image++) {
	for (int c = 0; c < 3; c++) {
	  int row = image * 3 + c;
	  ans.set(0, row, images[image][c].get(x, y));
	  eq.set(0, row, image & (1<<c) ? 1.0 : 0.0);
	  eq.set(c + 1, row, 1.0);
	}
      }
      ret.set(x, y, solve_nnls(eq, ans).get(0, 0));
    }
  }
  return ret;
}

Matrix<float> read_solution(int lights, int strings) {
  Matrix<float> solution = read_matrix("solution");
  // Matrix<float> solution = tmp.invert_nnls();
  Matrix<float> tmp(1, 120 * 90);
  tmp += 1.0;
  tmp = solution * tmp;
  double sum = 0.0;
  for (int l = 0; l < lights; l++)
    for (int s = 0; s < strings; s++)
      sum += tmp.get(0, l + s * lights);
  return solution * (lights * strings / sum);
}

int main(int argc, char **argv) {
  test();
  int strings = atoi(argv[1]);
  int lights = atoi(argv[2]);
  TestStream stream;
  Matrix<float> r(120, 90);
  Matrix<float> g(120, 90);
  Matrix<float> b(120, 90);
  TeensyLightWallRelay relay("/dev/ttyACM0", 8);
  relay.Init();

  Matrix<float> solution = read_solution(strings, lights);

  while (1) {
    r.Reshape(120, 90);
    g.Reshape(120, 90);
    b.Reshape(120, 90);
    stream.read_frame(&r, &g, &b);
    r.Reshape(1, r.columns() * r.rows());
    g.Reshape(1, g.columns() * g.rows());
    b.Reshape(1, b.columns() * b.rows());
#if 1
    Matrix<float> r_lights = solution * r;
    Matrix<float> g_lights = solution * g;
    Matrix<float> b_lights = solution * b;
#else
    Matrix<float> r_lights = solve_nnls(tmp, r);
    Matrix<float> g_lights = solve_nnls(tmp, g);
    Matrix<float> b_lights = solve_nnls(tmp, b);
#endif
    
    for (int l = 0; l < lights; l++) {
      vector<LampColor> colors(strings);
      for (int s = 0; s < strings; s++) {
	fprintf(stderr, "(%f)", r_lights.get(0, l + s * lights));
	colors[s].intensity = 0xCC;
	colors[s].color[0] = scalecolor(r_lights.get(0, l + s * lights));
	colors[s].color[1] = scalecolor(g_lights.get(0, l + s * lights));
	colors[s].color[2] = scalecolor(b_lights.get(0, l + s * lights));
      }
      relay.SetPixels(l, colors);
    }
  }
}
